﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

namespace Warena
{
    public partial class Routine
    {
        public IEnumerable<WoWUnit> EnemyCasting10
        {
            get
            {

                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            !u.IsPet &&
                            IsAttackable(u) &&
                            (!u.IsFriendly || u.IsHostile) &&
                            u.InLineOfSpellSight &&
                            u.IsCasting &&
                            u.Distance <= 10 &&
                            u.CurrentCastTimeLeft.TotalMilliseconds < CastTimeLeft && 
                            (IsCastingCC(u) || u.IsCastingHealingSpell)
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyCastingAnything10
        {
            get
            {

                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            !u.IsPet &&
                            IsAttackable(u) &&
                            (!u.IsFriendly || u.IsHostile) &&
                            u.InLineOfSpellSight &&
                            u.IsCasting &&
                            u.Distance <= 10
                            //u.CurrentCastTimeLeft.TotalMilliseconds < CastTimeLeft
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyCastingOnMeDistance40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            IsAttackable(u) &&
                            !u.IsMe &&
                            u.IsTargetingMeOrPet &&
                            u.IsAlive &&
                            !u.IsPet &&
                            u.InLineOfSpellSight &&
                            u.IsCasting &&
                            u.Distance <= 40 && (
                            u.CastingSpell.Name == "Polymorph" ||
                            u.CastingSpell.Name == "Repentence" ||
                            u.CastingSpell.Name == "Cyclone" ||
                            u.CastingSpell.Name == "Fear" ||
                            u.CastingSpell.Name == "Hex"
                            )
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWPlayer> EnemyCastingCC40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            IsAttackable(u) &&
                            !u.IsMe &&
                            !u.IsPet &&
                            u.InLineOfSpellSight &&
                            u.IsCasting &&
                            u.Distance <= 40 && (
                            u.CastingSpell.Name == "Polymorph" ||
                            u.CastingSpell.Name == "Cyclone" ||
                            u.CastingSpell.Name == "Fear" ||
                            u.CastingSpell.Name == "Hex"
                            )
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyCCInMelee
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            IsAttackable(u) &&
                            u.IsWithinMeleeRange &&
                            !IsInCC(u)
                            );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> FacingEnemyCCInMelee
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            IsAttackable(u) &&
                            !StyxWoW.Me.IsFacing(u) &&
                            u.IsWithinMeleeRange &&
                            !IsInCC(u)
                            );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWPlayer> EnemyDistance8
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            !u.IsFriendly || u.IsHostile &&
                            u.IsAlive &&
                            !u.IsMe &&
                            !u.IsPet &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 8
                        );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWPlayer> EnemyHealerTargets40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            IsAttackable(u) &&
                            !u.IsFriendly &&
                            !u.IsPet &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 40 &&
                            u.IsCastingHealingSpell
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyIsInvulnerable30
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            IsAttackable(u) &&
                            u.IsAlive &&
                            !u.IsMe &&
                            u.IsHostile &&
                            !u.IsPet &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 30 &&
                            u.HealthPercent <= 75 &&
                            (u.ActiveAuras.ContainsKey("Ice Block") ||
                            u.ActiveAuras.ContainsKey("Hand of Protection") ||
                            u.ActiveAuras.ContainsKey("Divine Shield"))
                     );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyTargetingMeDistance40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            IsAttackable(u) &&
                            !u.IsMe &&
                            !u.IsPet &&
                            u.IsTargetingMeOrPet &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 40
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyTotems30
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            IsAttackable(u.OwnedByUnit) &&
                            (
                                u.Name == "Spirit Link Totem" ||
                            //u.Name == "Grounding Totem" ||
                                u.Name == "Healing Stream Totem" ||
                                u.Name == "Tremor Totem"
                            ) &&
                            u.IsTotem &&
                            u.InLineOfSight &&
                            u.Distance <= 30
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> EnemyUnitsInMelee
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                        !u.IsDead &&
                        u.CanSelect &&
                        u.Attackable &&
                        !u.IsFriendly &&
                        u.IsWithinMeleeRange
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> SafeguardTargets25
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            !u.IsMe &&
                            u.IsFriendly &&
                            u.InLineOfSight &&
                            u.Distance <= 25
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWPlayer> FriendsNeedingHelp30
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            u.IsFriendly &&
                            !u.IsHostile &&
                            u.HealthPercent <= 25 &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 30
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWPlayer> FriendsNeedingHelp25
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            u.IsInMyPartyOrRaid &&
                            !u.IsMe &&
                            u.IsFriendly &&
                            !u.IsHostile &&
                            u.HealthPercent <= 25 &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 25
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWPlayer> FriendsInScatterShot40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            u.IsInMyPartyOrRaid &&
                            !u.IsMe &&
                            u.HasAura("Scatter Shot") &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 40
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWUnit> FreezingTrap40
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(
                        u =>
                            IsAttackable(u.OwnedByUnit) &&
                            u.Entry == 1499 &&
                            u.InLineOfSight &&
                            u.Distance <= 40 &&
                            u.Distance >= 8
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWUnit>();
                }
            }
        }
        public IEnumerable<WoWUnit> FreezingTraps40
        {
            get
            {
                IEnumerable<WoWUnit> result;
                using (StyxWoW.Memory.AcquireFrame())
                {
                    result = (
                        from u in ObjectManager.GetObjectsOfType<WoWUnit>(false, false)
                        where u.Name == "Freezing Trap" && IsAttackable(u.OwnedByUnit) && u.Distance >= 8 && u.Distance <= 40 && u.InLineOfSpellSight
                        select u).ToList<WoWUnit>();
                }
                return result;
            }
        }
        public IEnumerable<WoWUnit> FriendsInCC
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            !u.IsMe &&
                            !u.IsPet &&
                            u.IsAlive &&
                            u.IsInMyPartyOrRaid &&
                            IsInCC(u)
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        public IEnumerable<WoWPlayer> Friends20
        {
            get
            {
                try
                {
                    var ret = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                        u =>
                            u.IsAlive &&
                            !u.IsMe &&
                            u.IsFriendly &&
                            u.InLineOfSpellSight &&
                            u.Distance <= 20
                    );

                    return ret.ToList();
                }
                catch (NullReferenceException)
                {
                    return new List<WoWPlayer>();
                }
            }
        }
        IEnumerable<WoWUnit> UnfriendlyUnits
        {
            get { return ObjectManager.GetObjectsOfType<WoWUnit>(true, false).Where(u => !u.IsDead && u.CanSelect && u.Attackable && !u.IsFriendly && u.IsWithinMeleeRange); }
        }

        public bool IsAttackable(WoWUnit unit)
        {
            if (unit == null)
                return false;

            if (!unit.CanSelect || !unit.Attackable)
                return false;

            if (unit.IsDead)
                return false;

            if (unit.IsOnTransport)
                return false;

            if (unit.IsHostile || !unit.IsFriendly)
                return true;

            return true;
        }
        public bool CanMeleeInterrupt(WoWUnit unit)
        {
            if (unit != null)
            {
                if (
                    !IsAttackable(unit.CurrentTarget)
                    )
                    return false;

                if (
                    unit.IsHostile && 
                    unit.IsWithinMeleeRange && 
                    unit.IsCasting && 
                    unit.CanInterruptCurrentSpellCast
                    )
                    return true;

            }
            return false;
        }
        public bool IsCrowdControlled(WoWUnit unit)
        {
            if (unit != null)
            {
                Dictionary<string, WoWAura>.ValueCollection auras = unit.Auras.Values;

                return auras.Any(a =>
                           a.IsHarmful && (
                           a.Spell.Mechanic == WoWSpellMechanic.Banished ||
                           a.Spell.Mechanic == WoWSpellMechanic.Disoriented ||
                           a.Spell.Mechanic == WoWSpellMechanic.Charmed ||
                           a.Spell.Mechanic == WoWSpellMechanic.Horrified ||
                           a.Spell.Mechanic == WoWSpellMechanic.Incapacitated ||
                           a.Spell.Mechanic == WoWSpellMechanic.Polymorphed ||
                           a.Spell.Mechanic == WoWSpellMechanic.Sapped ||
                           a.Spell.Mechanic == WoWSpellMechanic.Shackled ||
                           a.Spell.Mechanic == WoWSpellMechanic.Asleep ||
                           a.Spell.Mechanic == WoWSpellMechanic.Frozen ||
                           a.Spell.Mechanic == WoWSpellMechanic.Invulnerable ||
                           a.Spell.Mechanic == WoWSpellMechanic.Invulnerable2 ||
                           a.Spell.Mechanic == WoWSpellMechanic.Turned ||
                           a.Spell.Mechanic == WoWSpellMechanic.Fleeing ||

                           // Really want to ignore hexed mobs.
                           a.Spell.Name == "Hex"));
            }
            return false;
        }
        public bool IsImmune(WoWUnit unit)
        {
            if (unit != null)
            {
                if (unit.ActiveAuras.ContainsKey("Ice Block") || 
                    unit.ActiveAuras.ContainsKey("Hand of Protection") ||
                    unit.ActiveAuras.ContainsKey("Divine Shield") || 
                    unit.ActiveAuras.ContainsKey("Deterrence") || 
                    unit.ActiveAuras.ContainsKey("Cyclone"))
                    return true;
            }
            return false;
        }
        public bool IsInCC(WoWUnit unit)
        {
            if (unit != null)
            {
                    if (unit.IsHostile &&
                       (unit.ActiveAuras.ContainsKey("Repentance") ||
                        unit.ActiveAuras.ContainsKey("Blinding Light") ||
                        unit.ActiveAuras.ContainsKey("Intimidating Shout") ||
                        unit.ActiveAuras.ContainsKey("Freezing Trap")))
                        return true;
            }
            return false;
        }
        public bool IsMovementImpaired(WoWUnit unit)
        {
            if (unit != null)
            {
                Dictionary<string, WoWAura>.ValueCollection auras = unit.Auras.Values;

                return auras.Any(a =>
                           a.IsHarmful && (a.Spell.Mechanic == WoWSpellMechanic.Snared || a.Spell.Mechanic == WoWSpellMechanic.Rooted));
            }
            return false;
        }
        public bool EnemyCastingCCOnFriendDistance20
        {
            get
            {
                try
                {
                    using (StyxWoW.Memory.AcquireFrame())
                    {

                        var retEnemy = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                                u =>
                                   IsAttackable(u) &&
                                   //should we exclude me?
                                   !u.IsMe &&
                                   !u.IsTargetingMeOrPet &&
                                   u.IsAlive &&
                                   !u.IsPet &&
                                   u.Distance <= 60 &&
                                   u.IsCasting &&
                                   u.CurrentCastTimeLeft.TotalMilliseconds < CastTimeLeft &&
                                   IsCastingCC(u)
                             );

                        var retFriends = ObjectManager.GetObjectsOfType<WoWPlayer>(true, false).Where(
                                u =>
                                   !u.IsMe &&
                                   u.IsAlive &&
                                   u.IsInMyPartyOrRaid &&
                                   !u.IsPet &&
                                   u.Distance <= 20
                                );

                        var enemyCastingCC = retEnemy.ToList().GetEnumerator();
                        var Friends = retFriends.ToList();

                        while (enemyCastingCC.MoveNext())
                        {
                            WoWPlayer u = enemyCastingCC.Current;
                        
                            if (Friends.Contains(u.CurrentTarget))
                                return true;
                        }

                    return false;

                    }
                }
                    catch (NullReferenceException)
                    {
                        return false;
                    }

            }
        }
        public bool IsCastingCC(WoWUnit u)
        {
            if (u.IsCasting &&
                u.CastingSpell.Name == "Fear" ||
                u.CastingSpell.Name == "Polymorph" ||
                u.CastingSpell.Name == "Cyclone" ||
                u.CastingSpell.Name == "Repentence" ||
                u.CastingSpell.Name == "Hex")
                return true;
            return false;
        }


        public bool ShouldCastMassReflect
        {
            get
            {
                HashSet<string> ccNames = new HashSet<string>()
            {
                "Repentence", "Cyclone", "Entangling Roots", "Fear", "Hex", "Polymorph"
            };

                IEnumerable<WoWPlayer> nearbyEnemiesCastingCrowdControl =
                ObjectManager.GetObjectsOfType<WoWPlayer>(true, false)
                .Where(p => p.IsCasting
                    && p.Distance <= 60
                    && p.IsHostile
                    && ccNames.Contains(p.CastingSpell.Name));

                IEnumerable<WoWPlayer> nearbyFriendlies =
                new HashSet<WoWPlayer>(ObjectManager.GetObjectsOfType<WoWPlayer>(true, false)
                .Where(p => p.IsFriendly
                    && p.Distance <= 20
                    && p.IsAlive));

                return nearbyEnemiesCastingCrowdControl
                .Any(e => nearbyFriendlies.Contains(e.CurrentTarget));
            }
        }
        


    


    }
}
